=== modified file 'bzrlib/builtins.py'
--- bzrlib/builtins.py	2007-07-31 05:10:59 +0000
+++ bzrlib/builtins.py	2007-07-31 11:20:50 +0000
@@ -581,7 +581,6 @@
     def run(self, location=None, remember=False, overwrite=False,
             revision=None, verbose=False,
             directory=None):
-        from bzrlib.tag import _merge_tags_if_possible
         # FIXME: too much stuff is in the command class
         revision_id = None
         mergeable = None
@@ -595,7 +594,7 @@
             branch_to = Branch.open_containing(directory)[0]
 
         if location is not None:
-            mergeable, location_transport = _get_bundle_helper(location)
+            mergeable, location_transport = _get_mergeable_helper(location)
 
         stored_loc = branch_to.get_parent()
         if location is None:
@@ -2639,135 +2638,174 @@
             directory=None,
             ):
         from bzrlib.tag import _merge_tags_if_possible
+        # This is actually a branch (or merge-directive) *location*.
+        location = branch
+        del branch
+
         if merge_type is None:
             merge_type = _mod_merge.Merge3Merger
 
         if directory is None: directory = u'.'
-        # XXX: jam 20070225 WorkingTree should be locked before you extract its
-        #      inventory. Because merge is a mutating operation, it really
-        #      should be a lock_write() for the whole cmd_merge operation.
-        #      However, cmd_merge open's its own tree in _merge_helper, which
-        #      means if we lock here, the later lock_write() will always block.
-        #      Either the merge helper code should be updated to take a tree,
-        #      (What about tree.merge_from_branch?)
+        possible_transports = []
+        merger = None
+        allow_pending = True
+        verified = 'inapplicable'
         tree = WorkingTree.open_containing(directory)[0]
         change_reporter = delta._ChangeReporter(
             unversioned_filter=tree.is_ignored)
-
-        other_transport = None
-        other_revision_id = None
-        base_revision_id = None
-        possible_transports = []
-
-        if branch is not None:
-            mergeable, other_transport = _get_bundle_helper(branch)
-            if mergeable:
-                if revision is not None:
-                    raise errors.BzrCommandError(
-                        'Cannot use -r with merge directives or bundles')
-                mergeable.install_revisions(tree.branch.repository)
-                base_revision_id, other_revision_id, verified =\
-                    mergeable.get_merge_request(tree.branch.repository)
-                if (base_revision_id != _mod_revision.NULL_REVISION and
-                    tree.branch.repository.get_graph().is_ancestor(
-                    base_revision_id, tree.branch.last_revision())):
-                    base_revision_id = None
-                other_branch = None
-                path = ''
-                other = None
-                base = None
-            possible_transports.append(other_transport)
-
-        if other_revision_id is None:
-            verified = 'inapplicable'
-            if revision is None \
-                    or len(revision) < 1 or revision[0].needs_branch():
-                branch = self._get_remembered_parent(tree, branch,
-                    'Merging from')
-
-            if revision is None or len(revision) < 1:
-                if uncommitted:
-                    base = [branch, -1]
-                    other = [branch, None]
-                else:
-                    base = [None, None]
-                    other = [branch, -1]
-                other_branch, path = Branch.open_containing(branch,
-                                                            possible_transports)
-            else:
-                if uncommitted:
+        cleanups = []
+        try:
+            pb = ui.ui_factory.nested_progress_bar()
+            cleanups.append(pb.finished)
+            tree.lock_write()
+            cleanups.append(tree.unlock)
+            if location is not None:
+                mergeable, other_transport = _get_mergeable_helper(location)
+                if mergeable:
+                    if uncommitted:
+                        raise errors.BzrCommandError('Cannot use --uncommitted'
+                            ' with bundles or merge directives.')
+
+                    if revision is not None:
+                        raise errors.BzrCommandError(
+                            'Cannot use -r with merge directives or bundles')
+                    merger, verified = _mod_merge.Merger.from_mergeable(tree,
+                       mergeable, pb)
+                possible_transports.append(other_transport)
+
+            if merger is None and uncommitted:
+                if revision is not None and len(revision) > 0:
                     raise errors.BzrCommandError('Cannot use --uncommitted and'
                         ' --revision at the same time.')
-                branch = revision[0].get_branch() or branch
-                if len(revision) == 1:
-                    base = [None, None]
-                    other_branch, path = Branch.open_containing(
-                        branch, possible_transports)
-                    revno = revision[0].in_history(other_branch).revno
-                    other = [branch, revno]
-                else:
-                    assert len(revision) == 2
-                    if None in revision:
-                        raise errors.BzrCommandError(
-                            "Merge doesn't permit empty revision specifier.")
-                    base_branch, path = Branch.open_containing(
-                        branch, possible_transports)
-                    branch1 = revision[1].get_branch() or branch
-                    other_branch, path1 = Branch.open_containing(
-                        branch1, possible_transports)
-                    if revision[0].get_branch() is not None:
-                        # then path was obtained from it, and is None.
-                        path = path1
-
-                    base = [branch, revision[0].in_history(base_branch).revno]
-                    other = [branch1,
-                             revision[1].in_history(other_branch).revno]
-
+                location = self._select_branch_location(tree, location)[0]
+                other_tree, other_path = WorkingTree.open_containing(location)
+                merger = _mod_merge.Merger.from_uncommitted(tree, other_tree,
+                    pb)
+                allow_pending = False
+
+            if merger is None:
+                merger, allow_pending = self._get_merger_from_branch(tree,
+                    location, revision, remember, possible_transports, pb)
+
+            merger.merge_type = merge_type
+            merger.reprocess = reprocess
+            merger.show_base = show_base
+            merger.change_reporter = change_reporter
+            self.sanity_check_merger(merger)
+            if (merger.base_rev_id == merger.other_rev_id and
+                merger.other_rev_id != None):
+                note('Nothing to do.')
+                return 0
+            if pull:
+                if merger.interesting_files is not None:
+                    raise BzrCommandError('Cannot pull individual files')
+                if (merger.base_rev_id == tree.last_revision()):
+                    result = tree.pull(merger.other_branch, False,
+                                       merger.other_rev_id)
+                    result.report(self.outf)
+                    return 0
+            merger.check_basis(not force)
+            conflict_count = merger.do_merge()
+            if allow_pending:
+                merger.set_pending()
+            if verified == 'failed':
+                warning('Preview patch does not match changes')
+            if conflict_count != 0:
+                return 1
+            else:
+                return 0
+        finally:
+            for cleanup in reversed(cleanups):
+                cleanup()
+
+    def sanity_check_merger(self, merger):
+        if (merger.show_base and
+            not merger.merge_type is _mod_merge.Merge3Merger):
+            raise errors.BzrCommandError("Show-base is not supported for this"
+                                         " merge type. %s" % merge_type)
+        if merger.reprocess and not merger.merge_type.supports_reprocess:
+            raise errors.BzrCommandError("Conflict reduction is not supported"
+                                         " for merge type %s." % merge_type)
+        if merger.reprocess and merger.show_base:
+            raise errors.BzrCommandError("Cannot do conflict reduction and"
+                                         " show base.")
+
+    def _get_merger_from_branch(self, tree, location, revision, remember,
+                                possible_transports, pb):
+        """Produce a merger from a location, assuming it refers to a branch."""
+        from bzrlib.tag import _merge_tags_if_possible
+        assert revision is None or len(revision) < 3
+        # find the branch locations
+        other_loc, location = self._select_branch_location(tree, location,
+            revision, -1)
+        if revision is not None and len(revision) == 2:
+            base_loc, location = self._select_branch_location(tree, location,
+                                                              revision, 0)
+        else:
+            base_loc = other_loc
+        # Open the branches
+        other_branch, other_path = Branch.open_containing(other_loc,
+            possible_transports)
+        if base_loc == other_loc:
+            base_branch = other_branch
+        else:
+            base_branch, base_path = Branch.open_containing(base_loc,
+                possible_transports)
+        # Find the revision ids
+        if revision is None or len(revision) < 1 or revision[-1] is None:
+            other_revision_id = _mod_revision.ensure_null(
+                other_branch.last_revision())
+        else:
+            other_revision_id = \
+                _mod_revision.ensure_null(
+                    revision[-1].in_history(other_branch).rev_id)
+        if (revision is not None and len(revision) == 2
+            and revision[0] is not None):
+            base_revision_id = \
+                _mod_revision.ensure_null(
+                    revision[0].in_history(base_branch).rev_id)
+        else:
+            base_revision_id = None
         # Remember where we merge from
         if ((tree.branch.get_parent() is None or remember) and
             other_branch is not None):
             tree.branch.set_parent(other_branch.base)
-
-        # pull tags now... it's a bit inconsistent to do it ahead of copying
-        # the history but that's done inside the merge code
-        if other_branch is not None:
-            _merge_tags_if_possible(other_branch, tree.branch)
-
-        if path != "":
-            interesting_files = [path]
+        _merge_tags_if_possible(other_branch, tree.branch)
+        merger = _mod_merge.Merger.from_revision_ids(pb, tree,
+            other_revision_id, base_revision_id, other_branch, base_branch)
+        if other_path != '':
+            allow_pending = False
+            merger.interesting_files = [other_path]
         else:
-            interesting_files = None
-        pb = ui.ui_factory.nested_progress_bar()
-        try:
-            try:
-                conflict_count = _merge_helper(
-                    other, base, other_rev_id=other_revision_id,
-                    base_rev_id=base_revision_id,
-                    check_clean=(not force),
-                    merge_type=merge_type,
-                    reprocess=reprocess,
-                    show_base=show_base,
-                    pull=pull,
-                    this_dir=directory,
-                    pb=pb, file_list=interesting_files,
-                    change_reporter=change_reporter,
-                    possible_transports=possible_transports)
-            finally:
-                pb.finished()
-            if verified == 'failed':
-                warning('Preview patch does not match changes')
-            if conflict_count != 0:
-                return 1
-            else:
-                return 0
-        except errors.AmbiguousBase, e:
-            m = ("sorry, bzr can't determine the right merge base yet\n"
-                 "candidates are:\n  "
-                 + "\n  ".join(e.bases)
-                 + "\n"
-                 "please specify an explicit base with -r,\n"
-                 "and (if you want) report this to the bzr developers\n")
-            log_error(m)
+            allow_pending = True
+        return merger, allow_pending
+
+    def _select_branch_location(self, tree, location, revision=None,
+                                index=None):
+        """Select a branch location, according to possible inputs.
+
+        If provided, branches from ``revision`` are preferred.  (Both
+        ``revision`` and ``index`` must be supplied.)
+
+        Otherwise, the ``location`` parameter is used.  If it is None, then the
+        ``parent`` location is used, and a note is printed.
+
+        :param tree: The working tree to select a branch for merging into
+        :param location: The location entered by the user
+        :param revision: The revision parameter to the command
+        :param index: The index to use for the revision parameter.  Negative
+            indices are permitted.
+        :return: (selected_location, default_location).  The default location
+            will be the user-entered location, if any, or else the remembered
+            location.
+        """
+        if (revision is not None and index is not None
+            and revision[index] is not None):
+            branch = revision[index].get_branch()
+            if branch is not None:
+                return branch, location
+        location = self._get_remembered_parent(tree, location, 'Merging from')
+        return location, location
 
     # TODO: move up to common parent; this isn't merge-specific anymore. 
     def _get_remembered_parent(self, tree, supplied_location, verb_string):
@@ -2781,8 +2819,10 @@
         mutter("%s", stored_location)
         if stored_location is None:
             raise errors.BzrCommandError("No location specified or remembered")
-        display_url = urlutils.unescape_for_display(stored_location, self.outf.encoding)
-        self.outf.write("%s remembered location %s\n" % (verb_string, display_url))
+        display_url = urlutils.unescape_for_display(stored_location,
+            self.outf.encoding)
+        self.outf.write("%s remembered location %s\n" % (verb_string,
+            display_url))
         return stored_location
 
 
@@ -3884,109 +3924,6 @@
             self.outf.write('%-20s %s\n' % (tag_name, target))
 
 
-# command-line interpretation helper for merge-related commands
-def _merge_helper(other_revision, base_revision,
-                  check_clean=True, ignore_zero=False,
-                  this_dir=None, backup_files=False,
-                  merge_type=None,
-                  file_list=None, show_base=False, reprocess=False,
-                  pull=False,
-                  pb=DummyProgress(),
-                  change_reporter=None,
-                  other_rev_id=None, base_rev_id=None,
-                  possible_transports=None):
-    """Merge changes into a tree.
-
-    base_revision
-        list(path, revno) Base for three-way merge.  
-        If [None, None] then a base will be automatically determined.
-    other_revision
-        list(path, revno) Other revision for three-way merge.
-    this_dir
-        Directory to merge changes into; '.' by default.
-    check_clean
-        If true, this_dir must have no uncommitted changes before the
-        merge begins.
-    ignore_zero - If true, suppress the "zero conflicts" message when 
-        there are no conflicts; should be set when doing something we expect
-        to complete perfectly.
-    file_list - If supplied, merge only changes to selected files.
-
-    All available ancestors of other_revision and base_revision are
-    automatically pulled into the branch.
-
-    The revno may be -1 to indicate the last revision on the branch, which is
-    the typical case.
-
-    This function is intended for use from the command line; programmatic
-    clients might prefer to call merge.merge_inner(), which has less magic 
-    behavior.
-    """
-    # Loading it late, so that we don't always have to import bzrlib.merge
-    if merge_type is None:
-        merge_type = _mod_merge.Merge3Merger
-    if this_dir is None:
-        this_dir = u'.'
-    this_tree = WorkingTree.open_containing(this_dir)[0]
-    if show_base and not merge_type is _mod_merge.Merge3Merger:
-        raise errors.BzrCommandError("Show-base is not supported for this merge"
-                                     " type. %s" % merge_type)
-    if reprocess and not merge_type.supports_reprocess:
-        raise errors.BzrCommandError("Conflict reduction is not supported for merge"
-                                     " type %s." % merge_type)
-    if reprocess and show_base:
-        raise errors.BzrCommandError("Cannot do conflict reduction and show base.")
-    # TODO: jam 20070226 We should really lock these trees earlier. However, we
-    #       only want to take out a lock_tree_write() if we don't have to pull
-    #       any ancestry. But merge might fetch ancestry in the middle, in
-    #       which case we would need a lock_write().
-    #       Because we cannot upgrade locks, for now we live with the fact that
-    #       the tree will be locked multiple times during a merge. (Maybe
-    #       read-only some of the time, but it means things will get read
-    #       multiple times.)
-    try:
-        merger = _mod_merge.Merger(this_tree.branch, this_tree=this_tree,
-                                   pb=pb, change_reporter=change_reporter)
-        merger.pp = ProgressPhase("Merge phase", 5, pb)
-        merger.pp.next_phase()
-        merger.check_basis(check_clean)
-        if other_rev_id is not None:
-            merger.set_other_revision(other_rev_id, this_tree.branch)
-        else:
-            merger.set_other(other_revision, possible_transports)
-        merger.pp.next_phase()
-        if base_rev_id is not None:
-            merger.set_base_revision(base_rev_id, this_tree.branch)
-        elif base_revision is not None:
-            merger.set_base(base_revision)
-        else:
-            merger.find_base()
-        if merger.base_rev_id == merger.other_rev_id:
-            note('Nothing to do.')
-            return 0
-        if file_list is None:
-            if pull and merger.base_rev_id == merger.this_rev_id:
-                # FIXME: deduplicate with pull
-                result = merger.this_tree.pull(merger.this_branch,
-                        False, merger.other_rev_id)
-                if result.old_revid == result.new_revid:
-                    note('No revisions to pull.')
-                else:
-                    note('Now on revision %d.' % result.new_revno)
-                return 0
-        merger.backup_files = backup_files
-        merger.merge_type = merge_type 
-        merger.set_interesting_files(file_list)
-        merger.show_base = show_base 
-        merger.reprocess = reprocess
-        conflicts = merger.do_merge()
-        if file_list is None:
-            merger.set_pending()
-    finally:
-        pb.clear()
-    return conflicts
-
-
 def _create_prefix(cur_transport):
     needed = [cur_transport]
     # Recurse upwards until we can create a directory successfully
@@ -4009,8 +3946,8 @@
         cur_transport.ensure_base()
 
 
-def _get_bundle_helper(location):
-    """Get a bundle if 'location' points to one.
+def _get_mergeable_helper(location):
+    """Get a merge directive or bundle if 'location' points to one.
 
     Try try to identify a bundle and returns its mergeable form. If it's not,
     we return the tried transport anyway so that it can reused to access the
@@ -4037,10 +3974,6 @@
     return mergeable, location_transport
 
 
-# Compatibility
-merge = _merge_helper
-
-
 # these get imported and then picked up by the scan for cmd_*
 # TODO: Some more consistent way to split command definitions across files;
 # we do need to load at least some information about them to know of 

=== modified file 'bzrlib/merge.py'
--- bzrlib/merge.py	2007-07-22 17:19:18 +0000
+++ bzrlib/merge.py	2007-07-31 11:20:50 +0000
@@ -89,6 +89,66 @@
         self.change_reporter = change_reporter
         self._cached_trees = {}
 
+    @staticmethod
+    def from_uncommitted(tree, other_tree, pb):
+        """Return a Merger for uncommitted changes in other_tree.
+
+        :param tree: The tree to merge into
+        :param other_tree: The tree to get uncommitted changes from
+        :param pb: A progress indicator
+        """
+        merger = Merger(tree.branch, other_tree, other_tree.basis_tree(), tree,
+                        pb)
+        merger.base_rev_id = merger.base_tree.get_revision_id()
+        merger.other_rev_id = None
+        return merger
+
+    @classmethod
+    def from_mergeable(klass, tree, mergeable, pb):
+        """Return a Merger for a bundle or merge directive.
+
+        :param tree: The tree to merge changes into
+        :param mergeable: A merge directive or bundle
+        :param pb: A progress indicator
+        """
+        mergeable.install_revisions(tree.branch.repository)
+        base_revision_id, other_revision_id, verified =\
+            mergeable.get_merge_request(tree.branch.repository)
+        if (base_revision_id != _mod_revision.NULL_REVISION and
+            tree.branch.repository.get_graph().is_ancestor(
+            base_revision_id, tree.branch.last_revision())):
+            base_revision_id = None
+        merger = klass.from_revision_ids(pb, tree, other_revision_id,
+                                         base_revision_id)
+        return merger, verified
+
+    @staticmethod
+    def from_revision_ids(pb, this, other, base=None, other_branch=None,
+                          base_branch=None):
+        """Return a Merger for revision-ids.
+
+        :param tree: The tree to merge changes into
+        :param other: The revision-id to use as OTHER
+        :param base: The revision-id to use as BASE.  If not specified, will
+            be auto-selected.
+        :param other_branch: A branch containing the other revision-id.  If
+            not supplied, this.branch is used.
+        :param base_branch: A branch containing the base revision-id.  If
+            not supplied, other_branch or this.branch will be used.
+        :param pb: A progress indicator
+        """
+        merger = Merger(this.branch, this_tree=this, pb=pb)
+        if other_branch is None:
+            other_branch = this.branch
+        merger.set_other_revision(other, other_branch)
+        if base is None:
+            merger.find_base()
+        else:
+            if base_branch is None:
+                base_branch = other_branch
+            merger.set_base_revision(base, base_branch)
+        return merger
+
     def revision_tree(self, revision_id, branch=None):
         if revision_id not in self._cached_trees:
             if branch is None:

=== modified file 'bzrlib/tests/blackbox/test_ancestry.py'
--- bzrlib/tests/blackbox/test_ancestry.py	2007-06-27 19:13:50 +0000
+++ bzrlib/tests/blackbox/test_ancestry.py	2007-07-25 22:03:13 +0000
@@ -16,7 +16,6 @@
 
 import os
 
-from bzrlib.builtins import merge
 from bzrlib.tests import TestCaseWithTransport
 from bzrlib.workingtree import WorkingTree
 from bzrlib.branch import Branch
@@ -35,7 +34,8 @@
         b_wt.commit('modified B/foo',rev_id='B1')
         open('A/foo', 'wb').write('000\n1111\n')
         a_wt.commit('modified A/foo',rev_id='A2')
-        merge(['B',-1],['B',1],this_dir='A')
+        a_wt.merge_from_branch(b_wt.branch, b_wt.last_revision(),
+            b_wt.branch.get_rev_id(1))
         a_wt.commit('merged B into A',rev_id='A3')
 
     def _check_ancestry(self, location='', result=None):

=== modified file 'bzrlib/tests/blackbox/test_merge.py'
--- bzrlib/tests/blackbox/test_merge.py	2007-07-27 05:17:06 +0000
+++ bzrlib/tests/blackbox/test_merge.py	2007-07-31 11:20:50 +0000
@@ -267,7 +267,7 @@
         self.pullable_branch()
         os.chdir('a')
         (out, err) = self.run_bzr('merge --pull ../b')
-        self.assertContainsRe(err, 'Now on revision 2\\.')
+        self.assertContainsRe(out, 'Now on revision 2\\.')
         tree_a = WorkingTree.open('.')
         self.assertEqual([self.id2], tree_a.get_parent_ids())
 
@@ -331,3 +331,30 @@
                              mangle_patch=True)
         err = self.run_bzr('merge -d target directive')[1]
         self.assertContainsRe(err, 'Preview patch does not match changes')
+
+    def test_merge_arbitrary(self):
+        target = self.make_branch_and_tree('target')
+        target.commit('empty')
+        # We need a revision that has no integer revno
+        branch_a = target.bzrdir.sprout('branch_a').open_workingtree()
+        self.build_tree(['branch_a/file1'])
+        branch_a.add('file1')
+        branch_a.commit('added file1', rev_id='rev2a')
+        branch_b = target.bzrdir.sprout('branch_b').open_workingtree()
+        self.build_tree(['branch_b/file2'])
+        branch_b.add('file2')
+        branch_b.commit('added file2', rev_id='rev2b')
+        branch_b.merge_from_branch(branch_a.branch)
+        self.failUnlessExists('branch_b/file1')
+        branch_b.commit('merged branch_a', rev_id='rev3b')
+
+        # It works if the revid has an interger revno
+        self.run_bzr('merge -d target -r revid:rev2a branch_a')
+        self.failUnlessExists('target/file1')
+        self.failIfExists('target/file2')
+        target.revert([])
+
+        # It should work if the revid has no integer revno
+        self.run_bzr('merge -d target -r revid:rev2a branch_b')
+        self.failUnlessExists('target/file1')
+        self.failIfExists('target/file2')

=== modified file 'bzrlib/tests/blackbox/test_status.py'
--- bzrlib/tests/blackbox/test_status.py	2007-07-23 14:27:42 +0000
+++ bzrlib/tests/blackbox/test_status.py	2007-07-25 22:03:13 +0000
@@ -34,7 +34,6 @@
     errors,
     )
 import bzrlib.branch
-from bzrlib.builtins import merge
 from bzrlib.osutils import pathjoin
 from bzrlib.revisionspec import RevisionSpec
 from bzrlib.status import show_tree_status
@@ -145,7 +144,7 @@
         b_2 = b_2_dir.open_branch()
         wt2 = b_2_dir.open_workingtree()
         wt.commit(u"\N{TIBETAN DIGIT TWO} Empty commit 2")
-        merge(["./branch", -1], [None, None], this_dir = './copy')
+        wt2.merge_from_branch(wt.branch)
         message = self.status_string(wt2)
         self.assertStartsWith(message, "pending merges:\n")
         self.assertEndsWith(message, "Empty commit 2\n")
@@ -153,7 +152,7 @@
         # must be long to make sure we see elipsis at the end
         wt.commit("Empty commit 3 " +
                    "blah blah blah blah " * 100)
-        merge(["./branch", -1], [None, None], this_dir = './copy')
+        wt2.merge_from_branch(wt.branch)
         message = self.status_string(wt2)
         self.assertStartsWith(message, "pending merges:\n")
         self.assert_("Empty commit 3" in message)

=== modified file 'bzrlib/tests/branch_implementations/test_bound_sftp.py'
--- bzrlib/tests/branch_implementations/test_bound_sftp.py	2007-04-13 07:39:17 +0000
+++ bzrlib/tests/branch_implementations/test_bound_sftp.py	2007-07-25 22:08:47 +0000
@@ -195,8 +195,6 @@
         self.assertEqual(['r@b-1'], sftp_b_newbase.revision_history())
 
     def test_bind_diverged(self):
-        from bzrlib.builtins import merge
-
         b_base, wt_child = self.create_branches()
 
         wt_child.branch.unbind()
@@ -286,8 +284,6 @@
         self.assertEqual(['r@b-1'], b_base.revision_history())
 
     def test_commit_after_merge(self):
-        from bzrlib.builtins import merge
-
         b_base, wt_child = self.create_branches()
 
         # We want merge to be able to be a local only

=== modified file 'bzrlib/tests/repository_implementations/test_fileid_involved.py'
--- bzrlib/tests/repository_implementations/test_fileid_involved.py	2007-03-28 22:54:43 +0000
+++ bzrlib/tests/repository_implementations/test_fileid_involved.py	2007-07-25 22:03:13 +0000
@@ -17,7 +17,6 @@
 import os
 import sys
 
-from bzrlib.builtins import merge
 from bzrlib.errors import IllegalPath, NonAsciiRevisionId
 from bzrlib.tests import TestSkipped
 from bzrlib.tests.repository_implementations.test_repository import TestCaseWithRepository

=== modified file 'bzrlib/tests/test_bundle.py'
--- bzrlib/tests/test_bundle.py	2007-07-19 18:16:11 +0000
+++ bzrlib/tests/test_bundle.py	2007-07-25 22:54:16 +0000
@@ -27,7 +27,6 @@
     revision as _mod_revision,
     treebuilder,
     )
-from bzrlib.builtins import _merge_helper
 from bzrlib.bzrdir import BzrDir
 from bzrlib.bundle.apply_bundle import install_bundle, merge_bundle
 from bzrlib.bundle.bundle_data import BundleTree
@@ -637,8 +636,7 @@
         self.assertEqualDiff(tree1_inv, tree2_inv)
         other.rename_one('sub/dir/nolastnewline.txt', 'sub/nolastnewline.txt')
         other.commit('rename file', rev_id='a@cset-0-6b')
-        _merge_helper([other.basedir, -1], [None, None],
-                      this_dir=self.tree1.basedir)
+        self.tree1.merge_from_branch(other.branch)
         self.tree1.commit(u'Merge', rev_id='a@cset-0-7',
                           verbose=False)
         bundle = self.get_valid_bundle('a@cset-0-6', 'a@cset-0-7')
@@ -740,8 +738,7 @@
         tt.create_file('file2', trans_id)
         tt.apply()
         other.commit('modify text in another tree', rev_id='a@lmod-0-2b')
-        _merge_helper([other.basedir, -1], [None, None],
-                      this_dir=self.tree1.basedir)
+        self.tree1.merge_from_branch(other.branch)
         self.tree1.commit(u'Merge', rev_id='a@lmod-0-3',
                           verbose=False)
         self.tree1.commit(u'Merge', rev_id='a@lmod-0-4')

=== modified file 'bzrlib/tests/test_fetch.py'
--- bzrlib/tests/test_fetch.py	2007-03-28 13:56:45 +0000
+++ bzrlib/tests/test_fetch.py	2007-07-25 22:03:13 +0000
@@ -21,11 +21,11 @@
 from bzrlib import (
     bzrdir,
     errors,
+    merge,
     repository,
     )
 from bzrlib.branch import Branch
 from bzrlib.bzrdir import BzrDir
-from bzrlib.builtins import merge
 import bzrlib.errors
 from bzrlib.repofmt import knitrepo
 from bzrlib.tests import TestCaseWithTransport
@@ -179,8 +179,7 @@
         wt2 = self.make_branch_and_tree('br2')
         br2 = wt2.branch
         wt2.commit(message='rev 2-1', rev_id='2-1')
-        merge(other_revision=['br1', -1], base_revision=['br1', 0],
-              this_dir='br2')
+        wt2.merge_from_branch(br1, from_revision='null:')
         self._check_revs_present(br2)
 
     def test_merge_fetches(self):
@@ -191,9 +190,9 @@
         dir_2 = br1.bzrdir.sprout('br2')
         br2 = dir_2.open_branch()
         wt1.commit(message='rev 1-2', rev_id='1-2')
-        dir_2.open_workingtree().commit(message='rev 2-1', rev_id='2-1')
-        merge(other_revision=['br1', -1], base_revision=[None, None], 
-              this_dir='br2')
+        wt2 = dir_2.open_workingtree()
+        wt2.commit(message='rev 2-1', rev_id='2-1')
+        wt2.merge_from_branch(br1)
         self._check_revs_present(br2)
 
     def _check_revs_present(self, br2):
@@ -228,8 +227,8 @@
     def test_merge_fetches_file_history(self):
         """Merge brings across file histories"""
         br2 = Branch.open('br2')
-        merge(other_revision=['br1', -1], base_revision=[None, None], 
-              this_dir='br2')
+        br1 = Branch.open('br1')
+        wt2 = WorkingTree.open('br2').merge_from_branch(br1)
         for rev_id, text in [('1-2', 'original from 1\n'),
                              ('1-3', 'agreement\n'),
                              ('2-1', 'contents in 2\n'),

=== modified file 'bzrlib/tests/test_merge.py'
--- bzrlib/tests/test_merge.py	2007-07-21 15:32:37 +0000
+++ bzrlib/tests/test_merge.py	2007-07-25 22:54:16 +0000
@@ -19,11 +19,12 @@
 
 from bzrlib import (
     conflicts,
+    errors,
     merge as _mod_merge,
     option,
+    progress,
     )
 from bzrlib.branch import Branch
-from bzrlib.builtins import _merge_helper
 from bzrlib.conflicts import ConflictList, TextConflict
 from bzrlib.errors import UnrelatedBranches, NoCommits, BzrCommandError
 from bzrlib.merge import transform_tree, merge_inner
@@ -41,28 +42,29 @@
         wt = self.make_branch_and_tree('.')
         rev_a = wt.commit("lala!")
         self.assertEqual([rev_a], wt.get_parent_ids())
-        _merge_helper([u'.', -1], [None, None])
+        self.assertRaises(errors.PointlessMerge, wt.merge_from_branch,
+                          wt.branch)
         self.assertEqual([rev_a], wt.get_parent_ids())
+        return wt
 
     def test_undo(self):
         wt = self.make_branch_and_tree('.')
         wt.commit("lala!")
         wt.commit("haha!")
         wt.commit("blabla!")
-        _merge_helper([u'.', 2], [u'.', 1])
+        wt.merge_from_branch(wt.branch, wt.branch.get_rev_id(2),
+                             wt.branch.get_rev_id(1))
 
     def test_nocommits(self):
-        self.test_pending()
+        wt = self.test_pending()
         wt2 = self.make_branch_and_tree('branch2')
-        self.assertRaises(NoCommits, _merge_helper, ['branch2', -1],
-                          [None, None])
-        return wt2
+        self.assertRaises(NoCommits, wt.merge_from_branch, wt2.branch)
+        return wt, wt2
 
     def test_unrelated(self):
-        wt2 = self.test_nocommits()
+        wt, wt2 = self.test_nocommits()
         wt2.commit("blah")
-        self.assertRaises(UnrelatedBranches, _merge_helper, ['branch2', -1],
-                          [None, None])
+        self.assertRaises(UnrelatedBranches, wt.merge_from_branch, wt2.branch)
         return wt2
 
     def test_merge_one_file(self):
@@ -92,9 +94,7 @@
         br1.fetch(wt2.branch)
         # merge all of branch 2 into branch 1 even though they 
         # are not related.
-        self.assertRaises(BzrCommandError, _merge_helper, ['branch2', -1],
-                          ['branch2', 0], reprocess=True, show_base=True)
-        _merge_helper(['branch2', -1], ['branch2', 0], reprocess=True)
+        wt1.merge_from_branch(wt2.branch, wt2.last_revision(), 'null:')
         self.assertEqual([br1.last_revision(), wt2.branch.last_revision()],
             wt1.get_parent_ids())
         return (wt1, wt2.branch)
@@ -280,5 +280,8 @@
         tree_a.commit('commit 2')
         tree_b = tree_a.bzrdir.sprout('b').open_workingtree()
         tree_b.rename_one('file_1', 'renamed')
-        _merge_helper(['b', None], ['b', -1], this_dir='a')
+        merger = _mod_merge.Merger.from_uncommitted(tree_a, tree_b,
+                                                    progress.DummyProgress())
+        merger.merge_type = _mod_merge.Merge3Merger
+        merger.do_merge()
         self.assertEqual(tree_a.get_parent_ids(), [tree_b.last_revision()])

=== modified file 'bzrlib/tests/test_merge_core.py'
--- bzrlib/tests/test_merge_core.py	2007-07-17 18:28:49 +0000
+++ bzrlib/tests/test_merge_core.py	2007-07-31 03:57:42 +0000
@@ -21,17 +21,24 @@
 import bzrlib
 from bzrlib import (
     generate_ids,
+    merge_directive,
     osutils,
     )
-from bzrlib.builtins import merge
 from bzrlib.conflicts import ContentsConflict, TextConflict, PathConflict
+from bzrlib import errors
 from bzrlib.errors import (NotBranchError, NotVersionedError,
                            WorkingTreeNotRevision, BzrCommandError, NoDiff3)
-import bzrlib.inventory as inventory
-from bzrlib.merge import Merge3Merger, Diff3Merger, WeaveMerger
+from bzrlib import  inventory
+from bzrlib.merge import (
+    Merge3Merger,
+    Diff3Merger,
+    WeaveMerger,
+    Merger,
+    )
 from bzrlib.osutils import (file_kind, getcwd, pathjoin, rename,
                             sha_file,
                             )
+from bzrlib import progress
 from bzrlib.transform import TreeTransform
 from bzrlib.tests import TestCaseWithTransport, TestCase, TestSkipped
 from bzrlib.workingtree import WorkingTree
@@ -454,11 +461,7 @@
         merge_type = Merge3Merger
         base = [None, None]
         other = ("mary", -1)
-        self.assertRaises(BzrCommandError, merge, other, base, check_clean=True,
-                          merge_type=WeaveMerger, this_dir="original",
-                          show_base=True)
-        merge(other, base, check_clean=True, merge_type=merge_type,
-              this_dir="original")
+        tree.merge_from_branch(mary_tree.branch)
         self.assertEqual("John\n", open("original/file1", "rt").read())
         self.assertEqual("Mary\n", open("original/file2", "rt").read())
  
@@ -476,13 +479,13 @@
         file('b/file', 'wb').write('this contents contents\n')
         wtb = d_b.open_workingtree()
         wtb.commit('this revision', allow_pointless=False)
-        self.assertEqual(merge(['a', -1], [None, None], this_dir='b'), 1)
+        self.assertEqual(1, wtb.merge_from_branch(wta.branch))
         self.assert_(os.path.lexists('b/file.THIS'))
         self.assert_(os.path.lexists('b/file.BASE'))
         self.assert_(os.path.lexists('b/file.OTHER'))
         wtb.revert([])
-        self.assertEqual(merge(['a', -1], [None, None], this_dir='b', 
-                               check_clean=False, merge_type=WeaveMerger), 1)
+        self.assertEqual(1, wtb.merge_from_branch(wta.branch,
+                                                  merge_type=WeaveMerger))
         self.assert_(os.path.lexists('b/file'))
         self.assert_(os.path.lexists('b/file.THIS'))
         self.assert_(not os.path.lexists('b/file.BASE'))
@@ -500,7 +503,7 @@
         file('b/b_file', 'wb').write('contents\n')
         wtb.add('b_file')
         b_rev = wtb.commit('b_revision', allow_pointless=False)
-        merge(['b', -1], ['b', 0], this_dir='a')
+        wta.merge_from_branch(wtb.branch, b_rev, 'null:')
         self.assert_(os.path.lexists('a/b_file'))
         self.assertEqual([b_rev], wta.get_parent_ids()[1:])
 
@@ -516,7 +519,7 @@
         file('b/file', 'wb').write('contents\n')
         wtb.add('file')
         b_rev = wtb.commit('b_revision', allow_pointless=False)
-        merge(['b', -1], ['b', 0], this_dir='a')
+        wta.merge_from_branch(wtb.branch, b_rev, 'null:')
         self.assert_(os.path.lexists('a/file'))
         self.assert_(os.path.lexists('a/file.moved'))
         self.assertEqual([b_rev], wta.get_parent_ids()[1:])
@@ -532,7 +535,8 @@
         file('b/file', 'wb').write('changed contents\n')
         wtb = WorkingTree.open('b')
         wtb.commit('changed file', allow_pointless=False)
-        merge(['a', -1], ['a', 1], this_dir='b')
+        wtb.merge_from_branch(wta.branch, wta.branch.last_revision(),
+                              wta.branch.get_rev_id(1))
         self.failIf(os.path.lexists('b/file'))
 
     def test_merge_metadata_vs_deletion(self):
@@ -549,7 +553,7 @@
         self.assertEqual(a_wt.branch.revno(), 2)
         self.assertFalse(os.path.exists('a/file'))
         b_wt.commit('exec a')
-        merge(['b', -1], ['b', 0], this_dir='a')
+        a_wt.merge_from_branch(b_wt.branch, b_wt.last_revision(), 'null:')
         self.assert_(os.path.exists('a/file'))
 
     def test_merge_swapping_renames(self):
@@ -565,7 +569,8 @@
         b_wt.rename_one('deux','un')
         b_wt.rename_one('tmp','deux')
         b_wt.commit('r1', rev_id='r1')
-        self.assertEqual(0, merge(['b', -1], ['b', 1], this_dir='a'))
+        self.assertEqual(0, a_wt.merge_from_branch(b_wt.branch,
+            b_wt.branch.last_revision(), b_wt.branch.get_rev_id(1)))
         self.failUnlessExists('a/un')
         self.failUnless('a/deux')
         self.assertFalse(os.path.exists('a/tmp'))
@@ -584,7 +589,8 @@
         file('b/file', 'wb').write('THAT')
         b_wt.add('file')
         b_wt.commit('r2')
-        merge(['b', -1],['b', 1],this_dir='a')
+        a_wt.merge_from_branch(b_wt.branch, b_wt.branch.last_revision(),
+                               b_wt.branch.get_rev_id(1))
         self.assert_(os.path.exists('a/file'))
         self.assertEqual(file('a/file').read(),'THAT')
 
@@ -612,7 +618,8 @@
         file('b/foo', 'wb').write('B/FOO')
         b_wt.add('foo')
         b_wt.commit('moved foo to bar, added new foo')
-        merge(['b', -1],['b', 1],this_dir='a')
+        a_wt.merge_from_branch(b_wt.branch, b_wt.branch.last_revision(),
+                               b_wt.branch.get_rev_id(1))
 
     def test_merge_create_before_rename(self):
         """create before rename, target parents before children
@@ -640,7 +647,8 @@
         b_wt.add('bar')
         b_wt.rename_one('foo', 'bar/foo')
         b_wt.commit('created bar dir, moved foo into bar')
-        merge(['b', -1],['b', 1],this_dir='a')
+        a_wt.merge_from_branch(b_wt.branch, b_wt.branch.last_revision(),
+                               b_wt.branch.get_rev_id(1))
 
     def test_merge_rename_to_temp_before_delete(self):
         """rename to temp before delete, source children before parents
@@ -669,7 +677,8 @@
         os.rmdir('b/foo')
         b_wt.remove('foo')
         b_wt.commit('moved foo/bar to bar, deleted foo')
-        merge(['b', -1],['b', 1],this_dir='a')
+        a_wt.merge_from_branch(b_wt.branch, b_wt.branch.last_revision(),
+                               b_wt.branch.get_rev_id(1))
 
     def test_merge_delete_before_rename_to_temp(self):
         """delete before rename to temp
@@ -698,5 +707,53 @@
         b_wt.remove('foo')
         b_wt.rename_one('bar', 'foo')
         b_wt.commit('deleted foo, renamed bar to foo')
-        merge(['b', -1],['b', 1],this_dir='a')
-
+        a_wt.merge_from_branch(b_wt.branch, b_wt.branch.last_revision(),
+                               b_wt.branch.get_rev_id(1))
+
+
+class TestMerger(TestCaseWithTransport):
+
+    def set_up_trees(self):
+        this = self.make_branch_and_tree('this')
+        this.commit('rev1', rev_id='rev1')
+        other = this.bzrdir.sprout('other').open_workingtree()
+        this.commit('rev2a', rev_id='rev2a')
+        other.commit('rev2b', rev_id='rev2b')
+        return this, other
+
+    def test_from_revision_ids(self):
+        this, other = self.set_up_trees()
+        self.assertRaises(errors.RevisionNotPresent, Merger.from_revision_ids,
+                          progress.DummyProgress(), this, 'rev2b')
+        merger = Merger.from_revision_ids(progress.DummyProgress(), this,
+            'rev2b', other_branch=other.branch)
+        self.assertEqual('rev2b', merger.other_rev_id)
+        self.assertEqual('rev1', merger.base_rev_id)
+        merger = Merger.from_revision_ids(progress.DummyProgress(), this,
+            'rev2b', 'rev2a', other_branch=other.branch)
+        self.assertEqual('rev2a', merger.base_rev_id)
+
+    def test_from_uncommitted(self):
+        this, other = self.set_up_trees()
+        merger = Merger.from_uncommitted(this, other, progress.DummyProgress())
+        self.assertIs(other, merger.other_tree)
+        self.assertIs(None, merger.other_rev_id)
+        self.assertEqual('rev2b', merger.base_rev_id)
+
+    def test_from_mergeable(self):
+        this, other = self.set_up_trees()
+        other.commit('rev3', rev_id='rev3')
+        md = merge_directive.MergeDirective2.from_objects(
+            other.branch.repository, 'rev3', 0, 0, 'this')
+        merger, verified = Merger.from_mergeable(this, md,
+            progress.DummyProgress())
+        md.patch = None
+        merger, verified = Merger.from_mergeable(this, md,
+            progress.DummyProgress())
+        self.assertEqual('inapplicable', verified)
+        self.assertEqual('rev3', merger.other_rev_id)
+        self.assertEqual('rev1', merger.base_rev_id)
+        md.base_revision_id = 'rev2b'
+        merger, verified = Merger.from_mergeable(this, md,
+            progress.DummyProgress())
+        self.assertEqual('rev2b', merger.base_rev_id)

=== modified file 'bzrlib/tests/test_missing.py'
--- bzrlib/tests/test_missing.py	2007-05-24 12:19:25 +0000
+++ bzrlib/tests/test_missing.py	2007-07-25 22:03:13 +0000
@@ -17,7 +17,6 @@
 import os
 
 
-from bzrlib.builtins import merge
 from bzrlib.missing import (
     find_unmerged,
     iter_log_data,
@@ -52,7 +51,7 @@
         self.assertEqual(find_unmerged(original, puller), ([], []))
         self.assertEqual(find_unmerged(original, merger), ([(2, u'b'), 
                                                             (3, u'c')], []))
-        merge(['original', -1], [None, None], this_dir='merger')
+        merger_tree.merge_from_branch(original)
         self.assertEqual(find_unmerged(original, merger), ([(2, u'b'), 
                                                             (3, u'c')], []))
         merger_tree.commit('d', rev_id='d')

=== modified file 'bzrlib/tests/test_revisionnamespaces.py'
--- bzrlib/tests/test_revisionnamespaces.py	2007-03-12 16:29:47 +0000
+++ bzrlib/tests/test_revisionnamespaces.py	2007-07-25 22:03:13 +0000
@@ -24,7 +24,6 @@
     errors,
     repository,
     )
-from bzrlib.builtins import merge
 from bzrlib.tests import TestCase, TestCaseWithTransport
 from bzrlib.revisionspec import (
     RevisionSpec,

=== modified file 'bzrlib/tests/workingtree_implementations/test_merge_from_branch.py'
--- bzrlib/tests/workingtree_implementations/test_merge_from_branch.py	2007-06-19 14:49:06 +0000
+++ bzrlib/tests/workingtree_implementations/test_merge_from_branch.py	2007-07-25 21:25:00 +0000
@@ -19,7 +19,10 @@
 
 import os
 
-from bzrlib import errors
+from bzrlib import (
+    errors,
+    merge
+    )
 from bzrlib.tests.workingtree_implementations import TestCaseWithWorkingTree
 
 
@@ -77,3 +80,37 @@
         self.assertRaises(errors.NoCommits, tree_a.merge_from_branch,
                           tree_b.branch)
         tree_b.merge_from_branch(tree_a.branch)
+
+    def test_merge_base(self):
+        tree_a = self.make_branch_and_tree('tree_a')
+        self.build_tree_contents([('tree_a/file', 'text-a')])
+        tree_a.add('file')
+        tree_a.commit('added file', rev_id='rev_1')
+        tree_b = tree_a.bzrdir.sprout('tree_b').open_workingtree()
+        os.unlink('tree_a/file')
+        tree_a.commit('deleted file')
+        self.build_tree_contents([('tree_b/file', 'text-b')])
+        tree_b.commit('changed file')
+        self.assertRaises(errors.PointlessMerge, tree_a.merge_from_branch,
+            tree_b.branch, from_revision=tree_b.branch.last_revision())
+        tree_a.merge_from_branch(tree_b.branch, from_revision='rev_1')
+        tree_a.lock_read()
+        self.addCleanup(tree_a.unlock)
+        changes = list(tree_a._iter_changes(tree_a.basis_tree()))
+        self.assertEqual(1, len(changes))
+
+    def test_merge_type(self):
+        this = self.make_branch_and_tree('this')
+        self.build_tree_contents([('this/foo', 'foo')])
+        this.add('foo', 'foo-id')
+        this.commit('added foo')
+        other = this.bzrdir.sprout('other').open_workingtree()
+        self.build_tree_contents([('other/foo', 'bar')])
+        other.commit('content -> bar')
+        self.build_tree_contents([('this/foo', 'baz')])
+        this.commit('content -> baz')
+        class QuxMerge(merge.Merge3Merger):
+            def text_merge(self, file_id, trans_id):
+                self.tt.create_file('qux', trans_id)
+        this.merge_from_branch(other.branch, merge_type=QuxMerge)
+        self.assertEqual('qux', this.get_file_text('foo-id'))

=== modified file 'bzrlib/workingtree.py'
--- bzrlib/workingtree.py	2007-07-17 18:28:49 +0000
+++ bzrlib/workingtree.py	2007-07-25 21:25:00 +0000
@@ -800,7 +800,8 @@
         self._control_files.put(filename, my_file)
 
     @needs_write_lock # because merge pulls data into the branch.
-    def merge_from_branch(self, branch, to_revision=None):
+    def merge_from_branch(self, branch, to_revision=None, from_revision=None,
+        merge_type=None):
         """Merge from a branch into this working tree.
 
         :param branch: The branch to merge from.
@@ -831,11 +832,17 @@
                 merger.other_rev_id)
             merger.other_branch = branch
             merger.pp.next_phase()
-            merger.find_base()
+            if from_revision is None:
+                merger.find_base()
+            else:
+                merger.set_base_revision(from_revision, branch)
             if merger.base_rev_id == merger.other_rev_id:
                 raise errors.PointlessMerge
             merger.backup_files = False
-            merger.merge_type = Merge3Merger
+            if merge_type is None:
+                merger.merge_type = Merge3Merger
+            else:
+                merger.merge_type = merge_type
             merger.set_interesting_files(None)
             merger.show_base = False
             merger.reprocess = False

